#define _GNU_SOURCE

#include "dmn.h"
#include "match.h"

void
usage (char *app)
{
  printf ("\n%s - firewall log watch daemon - ", app);
  printf ("[v %s-%s]\nIan Jones - roux@speakeasy.org\n\n", VERSION, RELEASE);
  printf
    ("Usage:\n\t %s [-k] [-d] [-m] [-s] [-b<buffersize>] [-l<logfile>]\n\t\t  [-t<template>] [-e<mail>] [-]\n\n",
     app);
  printf ("Use without any options to print alerts to the console.\n\n");
  exit (1);
}


void
quit (int sig)
{
  extern int dmn;
  if (!dmn)
    exit (sig);
  writebuf ();
  unlink (PIDFILE);
  openlog ("firelogd", LOG_CONS | LOG_NDELAY | LOG_PID, LOG_DAEMON);
  syslog (LOG_INFO, "STOP firewall log daemon [v %s-%s] <roux@speakeasy.org>",
	  VERSION, RELEASE);
  closelog ();
  exit (sig);
}

int
readbuf (void)
{
  FILE *buffer;
  extern int buf_counter;
  extern int buffersize;
  extern struct pktlog *mailbuffer;
  if ((buffer = fopen (BUFFERFILE, "r")) != NULL)
    {
      buf_counter =
	fread (mailbuffer, sizeof (struct pktlog), buffersize, buffer);
      fclose (buffer);
      unlink (BUFFERFILE);
    }
  return buf_counter;
}

int
writebuf (void)
{
  FILE *buffer;
  extern int buf_counter;
  extern struct pktlog *mailbuffer;
  int wrote = 0;
  if ((buffer = fopen (BUFFERFILE, "w")) == NULL)
    {
      return 0;
    }
  else
    {
      if (buf_counter != 0)
	{
	  wrote =
	    fwrite (mailbuffer, sizeof (struct pktlog), buf_counter, buffer);
	}
    }
  fclose (buffer);
  return wrote;
}

int
fmt_prn (FILE * out, const struct pktlog *pkt, int preproc_c,
	 char **preproc_v, const int recurse)
{
  int printed = 0;
  int counter = 0;
  char outst[1024] = "";

  if (preproc_c)
    {
      for (counter = 0; counter < preproc_c; counter++)
	{
	  if (strcmp (preproc_v[counter], "recursed") == 0)
	    {
	      if (!recurse && pkt->ptype == PROTO_ICMP
		  && pkt->lvl4.pkt_ICMP.hastrigger == 1)
		{
		  fmt_prn (out, pkt->lvl4.pkt_ICMP.recursed, preproc_c,
			   preproc_v, RECURSE);
		  free (pkt->lvl4.pkt_ICMP.recursed);
		}
	      continue;
	    }
	  pkt_str (preproc_v[counter], outst, pkt);
	  fputs (outst, out);
	  strcpy (outst, "");
	}
      printed++;
    }
  else
    {
      if (pkt->ptype == PROTO_ICMP)
	{
	  printed =
	    (fprintf
	     (out,
	      "%s/%s-%s ICMP: %s %s %s\n%s %s\n%s\t%s\n%s\t\t%s -> %s\n\n",
	      pkt->context.time, pkt->context.day, pkt->context.month,
	      pkt->context.msg, pkt->context.in, pkt->context.out,
	      pkt->lvl4.pkt_ICMP.info, pkt->lvl4.pkt_ICMP.trigger,
	      pkt->lvl3.fflags, pkt->lvl3.r_srcip, pkt->lvl3.sflags,
	      pkt->lvl3.srcip, pkt->lvl3.dstip));
	}
      else if (pkt->ptype == PROTO_TCP)
	{
	  printed =
	    (fprintf
	     (out,
	      "%s/%s-%s TCP %s %s %s %s %s/%s\n%s\t%s(%s)\n%s\t\t%s:%s -> %s:%s\n\n",
	      pkt->context.time, pkt->context.day, pkt->context.month,
	      pkt->lvl4.pkt_TCP.tflags, pkt->context.msg, pkt->context.in,
	      pkt->context.out, pkt->lvl4.pkt_TCP.dstpt,
	      pkt->lvl4.pkt_TCP.r_dstpt, pkt->lvl3.fflags, pkt->lvl3.r_srcip,
	      pkt->lvl4.pkt_TCP.r_srcpt, pkt->lvl3.sflags, pkt->lvl3.srcip,
	      pkt->lvl4.pkt_TCP.srcpt,
	      pkt->lvl3.dstip, pkt->lvl4.pkt_TCP.dstpt));
	}
      else if (pkt->ptype == PROTO_UDP)
	{
	  printed =
	    (fprintf
	     (out,
	      "%s/%s-%s UDP: %s %s %s %s/%s\n%s\t%s(%s)\n%s\t\t%s:%s -> %s:%s\n\n",
	      pkt->context.time, pkt->context.day, pkt->context.month,
	      pkt->context.msg, pkt->context.in, pkt->context.out,
	      pkt->lvl4.pkt_UDP.dstpt, pkt->lvl4.pkt_UDP.r_dstpt,
	      pkt->lvl3.fflags, pkt->lvl3.r_srcip, pkt->lvl4.pkt_UDP.r_srcpt,
	      pkt->lvl3.sflags, pkt->lvl3.srcip, pkt->lvl4.pkt_UDP.srcpt,
	      pkt->lvl3.dstip, pkt->lvl4.pkt_UDP.dstpt));
	}
      else
	{
	  printed =
	    (fprintf
	     (out, "%s/%s-%s  %s: %s %s %s\n*\t%s**\t\t%s -> %s\n\n",
	      pkt->context.time, pkt->context.day, pkt->context.month,
	      pkt->proto, pkt->context.msg, pkt->context.in, pkt->context.out,
	      pkt->lvl3.r_srcip, pkt->lvl3.srcip, pkt->lvl3.dstip));
	}

    }
  return printed;
}

void
pkt_str (const char *inst, char *outst, const struct pktlog *pkt)
{

  if (strcmp (inst, "month") == 0)
    strcpy (outst, pkt->context.month);
  else if (strcmp (inst, "day") == 0)
    strcpy (outst, pkt->context.day);
  else if (strcmp (inst, "time") == 0)
    strcpy (outst, pkt->context.time);
  else if (strcmp (inst, "msg") == 0)
    strcpy (outst, pkt->context.msg);
  else if (strcmp (inst, "in") == 0)
    strcpy (outst, pkt->context.in);
  else if (strcmp (inst, "out") == 0)
    strcpy (outst, pkt->context.out);
  else if (strcmp (inst, "mac") == 0)
    strcpy (outst, pkt->context.mac);
  else if (strcmp (inst, "srcip") == 0)
    strcpy (outst, pkt->lvl3.srcip);
  else if (strcmp (inst, "r_srcip") == 0)
    strcpy (outst, pkt->lvl3.r_srcip);
  else if (strcmp (inst, "dstip") == 0)
    strcpy (outst, pkt->lvl3.dstip);
  else if (strcmp (inst, "r_dstip") == 0)
    strcpy (outst, pkt->lvl3.r_dstip);
  else if (strcmp (inst, "iplen") == 0)
    strcpy (outst, pkt->lvl3.iplen);
  else if (strcmp (inst, "tos") == 0)
    strcpy (outst, pkt->lvl3.tos);
  else if (strcmp (inst, "pflags") == 0)
    strcpy (outst, pkt->lvl3.pflags);
  else if (strcmp (inst, "sflags") == 0)
    strcpy (outst, pkt->lvl3.sflags);
  else if (strcmp (inst, "prec") == 0)
    strcpy (outst, pkt->lvl3.prec);
  else if (strcmp (inst, "ttl") == 0)
    strcpy (outst, pkt->lvl3.ttl);
  else if (strcmp (inst, "id") == 0)
    strcpy (outst, pkt->lvl3.id);
  else if (strcmp (inst, "frag") == 0)
    strcpy (outst, pkt->lvl3.frag);
  else if (strcmp (inst, "fflags") == 0)
    strcpy (outst, pkt->lvl3.fflags);
  else if (strcmp (inst, "type") == 0)
    {
      if (pkt->ptype != PROTO_ICMP)
	strcpy (outst, "");
      else
	strcpy (outst, pkt->lvl4.pkt_ICMP.type);
    }
  else if (strcmp (inst, "code") == 0)
    {
      if (pkt->ptype != PROTO_ICMP)
	strcpy (outst, "");
      else
	strcpy (outst, pkt->lvl4.pkt_ICMP.code);
    }
  else if (strcmp (inst, "info") == 0)
    {
      if (pkt->ptype != PROTO_ICMP)
	strcpy (outst, "");
      else
	strcpy (outst, pkt->lvl4.pkt_ICMP.info);
    }
  else if (strcmp (inst, "trigger") == 0)
    {
      if (pkt->ptype != PROTO_ICMP)
	strcpy (outst, "");
      else
	strcpy (outst, pkt->lvl4.pkt_ICMP.trigger);
    }
  else if (strcmp (inst, "srcpt") == 0)
    {
      if (pkt->ptype != PROTO_TCP && pkt->ptype != PROTO_UDP)
	strcpy (outst, "");
      else
	strcpy (outst,
		pkt->ptype == PROTO_TCP ?
		pkt->lvl4.pkt_TCP.srcpt : pkt->lvl4.pkt_UDP.srcpt);
    }
  else if (strcmp (inst, "r_srcpt") == 0)
    {
      if (pkt->ptype != PROTO_TCP && pkt->ptype != PROTO_UDP)
	strcpy (outst, "");
      else
	strcpy (outst,
		pkt->ptype == PROTO_TCP ?
		pkt->lvl4.pkt_TCP.r_srcpt : pkt->lvl4.pkt_UDP.r_srcpt);
    }
  else if (strcmp (inst, "dstpt") == 0)
    {
      if (pkt->ptype != PROTO_TCP && pkt->ptype != PROTO_UDP)
	strcpy (outst, "");
      else
	strcpy (outst,
		pkt->ptype == PROTO_TCP ?
		pkt->lvl4.pkt_TCP.dstpt : pkt->lvl4.pkt_UDP.dstpt);
    }
  else if (strcmp (inst, "r_dstpt") == 0)
    {
      if (pkt->ptype != PROTO_TCP && pkt->ptype != PROTO_UDP)
	strcpy (outst, "");
      else
	strcpy (outst,
		pkt->ptype == PROTO_TCP ?
		pkt->lvl4.pkt_TCP.r_dstpt : pkt->lvl4.pkt_UDP.r_dstpt);
    }
  else if (strcmp (inst, "window") == 0)
    {
      if (pkt->ptype != PROTO_TCP)
	strcpy (outst, "");
      else
	strcpy (outst, pkt->lvl4.pkt_TCP.window);
    }
  else if (strcmp (inst, "res") == 0)
    {
      if (pkt->ptype != PROTO_TCP)
	strcpy (outst, "");
      else
	strcpy (outst, pkt->lvl4.pkt_TCP.res);
    }
  else if (strcmp (inst, "flags") == 0)
    {
      if (pkt->ptype != PROTO_TCP)
	strcpy (outst, "");
      else
	strcpy (outst, pkt->lvl4.pkt_TCP.flags);
    }
  else if (strcmp (inst, "tflags") == 0)
    {
      if (pkt->ptype != PROTO_TCP)
	strcpy (outst, "");
      else
	strcpy (outst, pkt->lvl4.pkt_TCP.tflags);
    }
  else if (strcmp (inst, "ulen") == 0)
    {
      if (pkt->ptype != PROTO_UDP)
	strcpy (outst, "");
      else
	strcpy (outst, pkt->lvl4.pkt_UDP.ulen);
    }
  else if (strcmp (inst, "proto") == 0)
    strcpy (outst, pkt->proto);
  else if (strcmp (inst, "log") == 0)
    strcpy (outst, pkt->log);
  else if (strcmp (inst, "ltype") == 0)
    {
      if (pkt->ltype == MAT_TABLE)
	strcpy (outst, "iptables");
      else
	strcpy (outst, "ipchains");
    }
  else if (strcmp (inst, "nl") == 0)
    strcpy (outst, "\n");
  else if (strcmp (inst, "sp") == 0)
    strcpy (outst, " ");
  else if (strcmp (inst, "tab") == 0)
    strcpy (outst, "\t");
  else
    strcpy (outst, inst);
}
