/*	FIREWALL LOG DAEMON UTILITIES		*/

#define _GNU_SOURCE

#include "dmn.h"
#include "match.h"

int buf_counter = 0;
int buffersize = 0;
int tblrchn = 0;
int klr, dmn, tailer, stin, mixed;
int lookopt = USE_EXT_SVC;
int opterr = 0;
struct pktlog *mailbuffer;
char *logfile = NULL;

int
main (int argc, char **argv)
{
  regex_t re;
  struct sigaction catch_all;
  struct stat sstat;
  extern int buf_counter, buffersize, lookopt;
  extern int tblrchn;
  extern int klr, dmn, tailer, stin, mixed;
  extern struct pktlog *mailbuffer;
  extern int opterr;

  char **preproc_v = NULL;
  char *mailcmd = NULL;
  FILE *mailproc;
  FILE *pidfile;
  FILE *prep_file;
  char buf[64000];
  char *template = NULL;
  struct pktlog matched;
  int counter, isfifo;
  int pid;
  int preproc_c = 0;
  int optc;
  static char opts[] = "e:mskdb:l:t:-";

  catch_all.sa_handler = quit;
  sigfillset (&catch_all.sa_mask);
  catch_all.sa_flags = 0;

  isfifo = mixed = klr = dmn = stin = 0;
  tailer = 1;

  while ((optc = getopt (argc, argv, opts)) != -1)
    {
      switch (optc)
	{
	case 'k':
	  klr++;
	  break;
	case 's':
	  lookopt = USE_STD_SVC;
	  break;
	case 'd':
	  dmn++;
	  tailer = 0;
	  break;
	case 'm':
	  mixed++;
	  break;
	case 'e':
	  asprintf (&mailcmd, "%s%s", MAILCMD, optarg);
	  break;
	case 'b':
	  dmn++;
	  tailer = 0;
	  buffersize = atoi (optarg);
	  if (!(buffersize > 0))
	    usage (argv[0]);
	  break;
	case 'l':
	  asprintf (&logfile, "%s", optarg);
	  break;
	case 't':
	  asprintf (&template, "%s", optarg);
	  break;
	default:
	  usage (argv[0]);
	}
    }
  for (; optind < argc; optind++)
    {
      if (strcmp (argv[optind], "-") == 0)
	{
	  stin++;
	  tailer++;
	  dmn = 0;
	}
      else
	usage (argv[0]);
    }
  if (argc == 1)
    tailer++;

  if ((pidfile = fopen (PIDFILE, "r")) != NULL)
    {
      fscanf (pidfile, "%d", &pid);
      fclose (pidfile);
    }
  else
    pid = 0;

  if (pid)
    {
      if (klr)
	{
	  if ((kill (pid, SIGINT) == -1))
	    fprintf (stderr, "%s: Stale PID. Removing %s.\n",
		     strerror (errno), PIDFILE);
	  else
	    printf ("Killed %d.\n", pid);
	  unlink (PIDFILE);
	  exit (0);
	}
      else if (!stin)
	{
	  printf
	    ("Did you want stdin? If so use '-'\n %s already running - pid: %d, exiting!\n",
	     argv[0], pid);
	  exit (1);
	}
    }
  else
    {
      if (klr)
	{
	  printf ("Nothing to kill!\n");
	  exit (1);
	}
    }

  if (template == NULL)
    asprintf (&template, "%s", TEMPLATE);
  prep_file = (fopen (template, "r"));
  free (template);
  if (prep_file)
    {
      int incomment = 0;
      while (fscanf (prep_file, "%63999s", buf) != EOF)
	{
	  if (strcmp (buf, "startcomment") == 0)
	    incomment = 1;
	  if (strcmp (buf, "endcomment") == 0)
	    {
	      incomment = 0;
	      continue;
	    }
	  if (incomment)
	    continue;
	  preproc_c++;
	}
      preproc_v = calloc ((size_t) preproc_c + 1, sizeof (char *));
      rewind (prep_file);
      preproc_c = 0;
      while (fscanf (prep_file, "%63999s", buf) != EOF)
	{
	  if (strcmp (buf, "startcomment") == 0)
	    incomment = 1;
	  if (strcmp (buf, "endcomment") == 0)
	    {
	      incomment = 0;
	      continue;
	    }
	  if (incomment)
	    continue;
	  if (
	      (preproc_v[preproc_c] =
	       (char *) malloc ((size_t) strlen (buf) * sizeof (char) + 1)) ==
	      NULL)
	    exit (7);
	  strcpy (preproc_v[preproc_c++], buf);
	}
      fclose (prep_file);
    }
  if (!stin)
    {
      if (logfile == NULL)
	asprintf (&logfile, "%s", LOGFILE);
      if (freopen (logfile, "r", stdin) == NULL)
	{
	  fprintf (stderr,
		   "\n%s:\nYou need to create %s using the command:\n\t\t\t\t\t\t\"mkfifo %s\"\nOr specify the logfile using the \"-l<log>\" option.\nDon't forget to change /etc/syslog.conf\n\n",
		   strerror (errno), logfile, logfile);
	  exit (13);
	}
      free (logfile);
      if (dmn)
	{
	  int kid;
	  FILE *fp;
	  if (buffersize == 0)
	    buffersize = BUFFERSIZE;
	  if (mailcmd == NULL)
	    asprintf (&mailcmd, "%s%s", MAILCMD, MAILTARGET);
	  if ((mailbuffer =
	       (struct pktlog *) malloc (buffersize *
					 sizeof (struct pktlog))) == NULL)
	    exit (7);
	  readbuf ();
	  if ((kid = fork ()))
	    {
	      fp = fopen (PIDFILE, "w");
	      fprintf (fp, "%d", kid);
	      fclose (fp);
	      exit (0);
	    }
	  setsid ();
	  sigaction (SIGINT, &catch_all, NULL);
	  sigaction (SIGHUP, &catch_all, NULL);
	  sigaction (SIGTERM, &catch_all, NULL);
	  sigaction (SIGQUIT, &catch_all, NULL);
	  sigaction (SIGKILL, &catch_all, NULL);
	  openlog ("firelogd", LOG_CONS | LOG_NDELAY | LOG_PID, LOG_DAEMON);
	  syslog (LOG_INFO,
		  "START firewall log daemon [v %s-%s] <roux@speakeasy.org>",
		  VERSION, RELEASE);
	  closelog ();
	}
      fstat (fileno (stdin), &sstat);
      if (S_ISFIFO (sstat.st_mode))
	isfifo = 1;		/* Set up for syslog restarts if input is not a file */
    }
read_loop:			/* Yeah, yeah...I know. Patches are welcome. */
  while (fgets (buf, sizeof (buf), stdin))
    {
      memset (&matched, 0, sizeof (struct pktlog));
      if (tblrchn && mixed)
	{
	  regfree (&re);
	  tblrchn = 0;
	}
      if (!tblrchn)
	{
	  if (regcomp (&re, TABLE_PAT, REG_EXTENDED) != 0)
	    {
	      printf ("ERROR in REGEX compilation\n");
	      return (0);
	    }
	  if (match (buf, &re, &matched, MAT_TABLE, NO_RECURSE))
	    tblrchn = MAT_TABLE;
	  else
	    {
	      regfree (&re);
	      if (regcomp (&re, CHAIN_PAT, REG_EXTENDED) != 0)
		{
		  printf ("ERROR in REGEX compilation\n");
		  return (0);
		}
	      if (match (buf, &re, &matched, MAT_CHAIN, NO_RECURSE))
		tblrchn = MAT_CHAIN;
	      else
		{
		  regfree (&re);
		  continue;
		}
	    }
	}
      else if (!(match (buf, &re, &matched, tblrchn, NO_RECURSE)))
	continue;
      if (tailer)
	fmt_prn (stdout, &matched, preproc_c, preproc_v, NO_RECURSE);
      else
	{
	  mailbuffer[buf_counter++] = matched;
	  if (buf_counter >= buffersize)
	    {
	      fflush (stdout);
	      if ((mailproc = popen (mailcmd, "w")) != NULL)
		{
		  for (counter = 0; counter < buf_counter; counter++)
		    fmt_prn (mailproc, &mailbuffer[counter], preproc_c,
			     preproc_v, NO_RECURSE);
		  fflush (mailproc);
		  pclose (mailproc);
		}
	      else
		fprintf (stderr, "%s: %s\nin dmn.h didn't work\n",
			 strerror (errno), MAILCMD);
	      buf_counter = 0;
	      mailbuffer[0] = matched;
	    }
	}
    }
  if (isfifo)
    {
      sleep (1);		/* wait for syslog to get it up */
      clearerr (stdin);
      goto read_loop;		/* heterodoxy */
    }
  quit (0);
  return (0);
}
