#define _GNU_SOURCE

#include "match.h"
#include "dmn.h"

int
match (const char *string,
       const regex_t * re, struct pktlog *pkt, const int type,
       const int recurse)
{
  int status;
  int isicmp = 0;
  unsigned len;
  unsigned count;
  unsigned safety = 2048;
  char out[safety];
  char tmp[safety];
  char *tok;
  char *tokptr;
  size_t nmatch;
  regmatch_t *matches;

  struct protoent *newproto;

  nmatch = re->re_nsub;
  matches = calloc (nmatch + 1, sizeof (regmatch_t));
  if (matches == NULL)
    {
      printf ("calloc failed - match.c, line 26\n");
      exit (13);
    }
  status = regexec (re, string, nmatch + 1, matches, 0);
  if (status != 0)
    {
/*      printf ("Log Line:\n%s\n", string);
      regerror (status, re, tmp, sizeof tmp);
      printf ("%s: regerror\n", tmp);*/
      return (0);
    }
  if (strlen (string) >= sizeof pkt->log)
    return (0);
  strcpy (pkt->log, string);
  for (isicmp = 0; isicmp < strlen (pkt->log); isicmp++)	/* isicmp is just */
    {				/*   a handy int */
      if (pkt->log[isicmp] == '\n')
	pkt->log[isicmp] = 0;
    }
  if (type == MAT_TABLE)
    {
      pkt->ltype = MAT_TABLE;
      for (count = 1; count < nmatch; count++)
	{
	  len = matches[count].rm_eo - matches[count].rm_so;
	  if (len >= safety)
	    return (0);
	  strncpy (out, string + matches[count].rm_so, len);
	  out[len] = 0;
	  switch (count)
/* 	OK, if you are actually reading this, you are about to get mad at me.
 * 	Typically, you will follow the cases in some kind of order, either 
 * 	numeric or alphabetical. I chose not to do this because every case will
 * 	be true once, so there is no efficiency to be gained by minimizing
 * 	fruitless matches. Because I have to maintain this mess, I decided to
 * 	follow the order that the variables are declared in the struct pktlog
 * 	instead. Please forgive me if this offends you.			*/
	    {			/* But we always put protocol first :) */
	    case 17:		/* fall through */
	    case 20:		/* ICMP */
	      if (strcmp (out, "ICMP") != 0)
		break;
	      pkt->ptype = PROTO_ICMP;
	      if (strlen (out) >= sizeof pkt->proto)
		return (0);
	      strcpy (pkt->proto, out);
	      break;
	    case 24:
	      if (strcmp (out, "TCP") != 0)
		break;
	      pkt->ptype = PROTO_TCP;
	      if (strlen (out) >= sizeof pkt->proto)
		return (0);
	      strcpy (pkt->proto, out);
	      break;
	    case 30:
	      if (strcmp (out, "UDP") != 0)
		break;
	      pkt->ptype = PROTO_UDP;
	      if (strlen (out) >= sizeof pkt->proto)
		return (0);
	      strcpy (pkt->proto, out);
	      break;
	    case 34:
	      if (strcmp (out, "ESP") != 0)
		break;
	      pkt->ptype = PROTO_ESP;
	      if (strlen (out) >= sizeof pkt->proto)
		return (0);
	      strcpy (pkt->proto, out);
	      break;
	    case 36:
	      if (strcmp (out, "AH") != 0)
		break;
	      pkt->ptype = PROTO_AH;
	      if (strlen (out) >= sizeof pkt->proto)
		return (0);
	      strcpy (pkt->proto, out);
	      break;
	    case 38:
	      if (pkt->ptype)
		break;
	      pkt->ptype = PROTO_UNK;
	      if (strlen (out) >= sizeof pkt->proto)
		return (0);
	      strcpy (pkt->proto, out);
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_UNK.unkdata)
		return (0);
	      strcpy (pkt->lvl4.pkt_UNK.unkdata, out);
	      break;
	    case 1:
	      if (strlen (out) >= sizeof pkt->context.month)
		return (0);
	      strcpy (pkt->context.month, out);
	      break;
	    case 2:
	      if (strlen (out) >= sizeof pkt->context.day)
		return (0);
	      strcpy (pkt->context.day, out);
	      break;
	    case 3:
	      if (strlen (out) >= sizeof pkt->context.time)
		return (0);
	      strcpy (pkt->context.time, out);
	      break;
	    case 4:
	      if (strlen (out) >= sizeof pkt->context.msg)
		return (0);
	      strcpy (pkt->context.msg, out);
	      break;
	    case 5:
	      if (strlen (out) >= sizeof pkt->context.in)
		return (0);
	      strcpy (pkt->context.in, out);
	      break;
	    case 6:
	      if (strlen (out) >= sizeof pkt->context.out)
		return (0);
	      strcpy (pkt->context.out, out);
	      break;
	    case 7:
	      strncpy (pkt->context.mac, out, sizeof pkt->context.mac - 1);
	      pkt->context.mac[45] = 0;
	      break;
	    case 8:
	      if (strlen (out) >= sizeof pkt->lvl3.srcip)
		return (0);
	      strcpy (pkt->lvl3.srcip, out);
	      strcpy (pkt->lvl3.r_srcip, out);
	      break;
	    case 9:
	      if (strlen (out) >= sizeof pkt->lvl3.dstip)
		return (0);
	      strcpy (pkt->lvl3.dstip, out);
	      strcpy (pkt->lvl3.r_dstip, out);
	      break;
	    case 10:
	      if (strlen (out) >= sizeof pkt->lvl3.iplen)
		return (0);
	      strcpy (pkt->lvl3.iplen, out);
	      break;
	    case 11:
	      if (strlen (out) >= sizeof pkt->lvl3.tos)
		return (0);
	      strcpy (pkt->lvl3.tos, out);
	      sscanf (out, "0x%x", &isicmp);	/* isicmp just a handy int */
	      pkt->lvl3.sflags[5] = 0;	/* Null terminate string */
	      pkt->lvl3.sflags[4] = isicmp & 0x01 ? 'C' : '*';	/* CE bit (ECN) */
	      pkt->lvl3.sflags[3] = isicmp & 0x02 ? 'E' : '*';	/* ECT bit (ECN) */
	      pkt->lvl3.sflags[2] = isicmp & 0x04 ? 'R' : '*';	/* reliability */
	      pkt->lvl3.sflags[1] = isicmp & 0x08 ? 'T' : '*';	/* throughput */
	      pkt->lvl3.sflags[0] = isicmp & 0x10 ? 'D' : '*';	/* delay */
	      break;
	    case 12:
	      if (strlen (out) >= sizeof pkt->lvl3.prec)
		return (0);
	      strcpy (pkt->lvl3.prec, out);
	      sscanf (out, "0x%x", &isicmp);	/* isicmp just a handy int */
	      pkt->lvl3.pflags[3] = 0;	/* Null terminate string */
	      pkt->lvl3.pflags[2] = isicmp & 0x01 ? '3' : '*';	/* precedence */
	      pkt->lvl3.pflags[1] = isicmp & 0x02 ? '2' : '*';	/* bits */
	      pkt->lvl3.pflags[0] = isicmp & 0x04 ? '1' : '*';
	      break;
	    case 13:
	      if (strlen (out) >= sizeof pkt->lvl3.ttl)
		return (0);
	      strcpy (pkt->lvl3.ttl, out);
	      break;
	    case 14:
	      if (strlen (out) >= sizeof pkt->lvl3.id)
		return (0);
	      strcpy (pkt->lvl3.id, out);
	      break;
	    case 15:
	      if (strlen (out) >= sizeof pkt->lvl3.frag)
		return (0);
	      strcpy (pkt->lvl3.frag, out);
	      strcpy (pkt->lvl3.fflags, "***");
	      tokptr = out;
	      while ((tok = strsep (&tokptr, " ")) != NULL)
		{
		  switch (tok[0])
		    {
		    case 'C':
		      pkt->lvl3.fflags[0] = 'C';	/* CONGESTION?, harumph! */
		      break;
		    case 'D':
		      pkt->lvl3.fflags[1] = 'D';
		      break;
		    case 'M':
		      pkt->lvl3.fflags[2] = 'M';
		      break;
		    }
		}
	      break;
	    case 18:
	    case 21:
	      if (pkt->ptype != PROTO_ICMP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_ICMP.type)
		return (0);
	      if (isdigit (pkt->lvl4.pkt_ICMP.type[0]))
		break;
	      strcpy (pkt->lvl4.pkt_ICMP.type, out);
	      break;
	    case 19:
	    case 22:
	      if (pkt->ptype != PROTO_ICMP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_ICMP.code)
		return (0);
	      if (isdigit (pkt->lvl4.pkt_ICMP.code[0]))
		break;
	      strcpy (pkt->lvl4.pkt_ICMP.code, out);
	      break;
	    case 23:
	      if (pkt->ptype != PROTO_ICMP || strcmp (out, "") == 0)
		break;
	      snprintf (pkt->lvl4.pkt_ICMP.trigger, (size_t) 1024,
			"%s %s %s loghost kernel: TRIGGERING PACKET IN= OUT= %s",
			pkt->context.month, pkt->context.day,
			pkt->context.time, out + 1);
	      for (isicmp = 0; isicmp < strlen (pkt->lvl4.pkt_ICMP.trigger);
		   isicmp++)	/* isicmp is just */
		{		/*   a handy int */
		  if (pkt->lvl4.pkt_ICMP.trigger[isicmp] == ']')
		    pkt->lvl4.pkt_ICMP.trigger[isicmp] = 0;
		}
	      if (!recurse)	/* one level of recursion only !! */
		{
		  pkt->lvl4.pkt_ICMP.recursed =
		    (struct pktlog *) malloc (sizeof (struct pktlog));
		  if (pkt->lvl4.pkt_ICMP.recursed == NULL)
		    break;
		  memset (pkt->lvl4.pkt_ICMP.recursed, 0,
			  sizeof (struct pktlog));
		  if (match
		      (pkt->lvl4.pkt_ICMP.trigger, re,
		       pkt->lvl4.pkt_ICMP.recursed, MAT_TABLE, RECURSE))
		    pkt->lvl4.pkt_ICMP.hastrigger = 1;
		}
	      break;
	    case 25:
	      if (pkt->ptype != PROTO_TCP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_TCP.srcpt)
		return (0);
	      strcpy (pkt->lvl4.pkt_TCP.srcpt, out);
	      strcpy (pkt->lvl4.pkt_TCP.r_srcpt, out);
	      break;
	    case 26:
	      if (pkt->ptype != PROTO_TCP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_TCP.dstpt)
		return (0);
	      strcpy (pkt->lvl4.pkt_TCP.dstpt, out);
	      strcpy (pkt->lvl4.pkt_TCP.r_dstpt, out);
	      break;
	    case 27:
	      if (pkt->ptype != PROTO_TCP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_TCP.window)
		return (0);
	      strcpy (pkt->lvl4.pkt_TCP.window, out);
	      break;
	    case 28:
	      if (pkt->ptype != PROTO_TCP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_TCP.res)
		return (0);
	      strcpy (pkt->lvl4.pkt_TCP.res, out);
	      strcpy (pkt->lvl4.pkt_TCP.tflags, "********");
	      sscanf (out, "%x", &isicmp);	/* isicmp is just a handy int */
	      switch (isicmp & 0x03)
		{
		case 0x2:
		  pkt->lvl4.pkt_TCP.tflags[0] = '1';	/* reserved bits 1 & 2 */
		  break;
		case 0x1:
		  pkt->lvl4.pkt_TCP.tflags[1] = '2';
		  break;
		case 0x3:
		  pkt->lvl4.pkt_TCP.tflags[0] = '1';
		  pkt->lvl4.pkt_TCP.tflags[1] = '2';
		}
	      break;
	    case 29:		// URG ACK PSH RST SYN FIN
	      if (pkt->ptype != PROTO_TCP)
		break;
	      memset (tmp, 0, safety);
	      if (strlen (out) >= sizeof tmp)
		return (0);
	      strcpy (tmp, " ");
	      tokptr = out;
	      while ((tok = strsep (&tokptr, " ")) != NULL)
		{
		  if (strstr (tok, "URGP") != NULL)
		    continue;
		  if (isupper (tmp[strlen (tmp) - 1]))
		    sprintf (tmp, "%s %s", tmp, tok);
		  else
		    sprintf (tmp, "%s", tok);
		  switch (tok[0])
		    {
		    case 'U':
		      pkt->lvl4.pkt_TCP.tflags[2] = 'U';
		      break;
		    case 'A':
		      pkt->lvl4.pkt_TCP.tflags[3] = 'A';
		      break;
		    case 'P':
		      pkt->lvl4.pkt_TCP.tflags[4] = 'P';
		      break;
		    case 'R':
		      pkt->lvl4.pkt_TCP.tflags[5] = 'R';
		      break;
		    case 'S':
		      pkt->lvl4.pkt_TCP.tflags[6] = 'S';
		      break;
		    case 'F':
		      pkt->lvl4.pkt_TCP.tflags[7] = 'F';
		      break;
		    }
		}
	      if (strlen (tmp) >= sizeof pkt->lvl4.pkt_TCP.flags)
		return 0;
	      strcpy (pkt->lvl4.pkt_TCP.flags, tmp);
	      break;
	    case 31:
	      if (pkt->ptype != PROTO_UDP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_UDP.srcpt)
		return (0);
	      strcpy (pkt->lvl4.pkt_UDP.srcpt, out);
	      strcpy (pkt->lvl4.pkt_UDP.r_srcpt, out);
	      break;
	    case 32:
	      if (pkt->ptype != PROTO_UDP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_UDP.dstpt)
		return (0);
	      strcpy (pkt->lvl4.pkt_UDP.dstpt, out);
	      strcpy (pkt->lvl4.pkt_UDP.r_dstpt, out);
	      break;
	    case 33:
	      if (pkt->ptype != PROTO_UDP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_UDP.ulen)
		return (0);
	      strcpy (pkt->lvl4.pkt_UDP.ulen, out);
	      break;
	    case 35:
	      if (pkt->ptype != PROTO_ESP)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_ESP.espdata)
		return (0);
	      strcpy (pkt->lvl4.pkt_ESP.espdata, out);
	      break;
	    case 37:
	      if (pkt->ptype != PROTO_AH)
		break;
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_AH.ahdata)
		return (0);
	      strcpy (pkt->lvl4.pkt_AH.ahdata, out);
	      break;
	    default:
	      continue;
	    }
	}
    }
  else if (type == MAT_CHAIN)
    {
      pkt->ltype = MAT_CHAIN;
      for (count = 1; count < nmatch; count++)
	{
	  len = matches[count].rm_eo - matches[count].rm_so;
	  if (len >= safety)
	    return (0);
	  strncpy (out, string + matches[count].rm_so, len);
	  out[len] = 0;
	  switch (count)
	    {
	    case 7:
	      isicmp = atoi (out);
	      if (isicmp == 1)
		{
		  pkt->ptype = PROTO_ICMP;
		  strcpy (pkt->proto, "icmp");
		}
	      else if (isicmp == 6)
		{
		  pkt->ptype = PROTO_TCP;
		  strcpy (pkt->proto, "tcp");
		}
	      else if (isicmp == 17)
		{
		  pkt->ptype = PROTO_UDP;
		  strcpy (pkt->proto, "udp");
		}
	      else
		{
		  pkt->ptype = PROTO_UNK;
		  newproto = getprotobynumber (atoi (out));
		  if (strlen (newproto->p_name) >=
		      sizeof pkt->lvl4.pkt_UNK.unkdata)
		    return (0);
		  strcpy (pkt->lvl4.pkt_UNK.unkdata, newproto->p_name);
		  if (strlen (newproto->p_name) >= sizeof pkt->proto)
		    return (0);
		  strcpy (pkt->proto, out);
		}
	      break;
	    case 1:
	      if (strlen (out) >= sizeof pkt->context.month)
		return (0);
	      strcpy (pkt->context.month, out);
	      break;
	    case 2:
	      if (strlen (out) >= sizeof pkt->context.day)
		return (0);
	      strcpy (pkt->context.day, out);
	      break;
	    case 3:
	      if (strlen (out) >= sizeof pkt->context.time)
		return (0);
	      strcpy (pkt->context.time, out);
	      break;
	    case 4:		// chain name
	      strcpy (tmp, out);
	      break;
	    case 5:		// action (DENY...)
	      strcpy (pkt->context.msg, out);
	      strcat (pkt->context.msg, "/");
	      strcat (pkt->context.msg, tmp);
	      break;
	    case 6:
	      strcpy (pkt->context.in, out);
	      break;
	    case 18:		// rule number
	      strcat (pkt->context.msg, "-");
	      strcat (pkt->context.msg, out);
	      break;
	    case 8:
	      if (strlen (out) >= sizeof pkt->lvl3.srcip)
		return (0);
	      strcpy (pkt->lvl3.srcip, out);
	      strcpy (pkt->lvl3.r_srcip, out);
	      break;
	    case 9:
	      if (pkt->ptype == PROTO_ICMP)
		strcpy (pkt->lvl4.pkt_ICMP.type, out);
	      else if (pkt->ptype == PROTO_TCP)
		{
		  strcpy (pkt->lvl4.pkt_TCP.srcpt, out);
		  strcpy (pkt->lvl4.pkt_TCP.r_srcpt, out);
		}
	      else if (pkt->ptype == PROTO_UDP)
		{
		  strcpy (pkt->lvl4.pkt_UDP.srcpt, out);
		  strcpy (pkt->lvl4.pkt_UDP.r_srcpt, out);
		}
	      break;
	    case 10:
	      if (strlen (out) >= sizeof pkt->lvl3.dstip)
		return (0);
	      strcpy (pkt->lvl3.dstip, out);
	      strcpy (pkt->lvl3.r_dstip, out);
	      break;
	    case 11:
	      if (pkt->ptype == PROTO_ICMP)
		strcpy (pkt->lvl4.pkt_ICMP.code, out);
	      else if (pkt->ptype == PROTO_TCP)
		{
		  strcpy (pkt->lvl4.pkt_TCP.dstpt, out);
		  strcpy (pkt->lvl4.pkt_TCP.r_dstpt, out);
		}
	      else if (pkt->ptype == PROTO_UDP)
		{
		  strcpy (pkt->lvl4.pkt_UDP.dstpt, out);
		  strcpy (pkt->lvl4.pkt_UDP.r_dstpt, out);
		}
	      break;
	    case 12:
	      if (strlen (out) >= sizeof pkt->lvl3.iplen)
		return (0);
	      strcpy (pkt->lvl3.iplen, out);
	      break;
	    case 13:
	      if (strlen (out) >= sizeof pkt->lvl3.tos)
		return (0);
	      strcpy (pkt->lvl3.tos, out);
	      sscanf (out, "0x%x", &isicmp);	/* isicmp just a handy int */
	      sprintf (pkt->lvl3.prec, "0x%02X", isicmp >> 5);
	      pkt->lvl3.sflags[5] = 0;	/* Null terminate TOS flags */
	      pkt->lvl3.sflags[4] = isicmp & 0x01 ? 'C' : '*';	/* CE bit (ECN) */
	      pkt->lvl3.sflags[3] = isicmp & 0x02 ? 'E' : '*';	/* ECT bit (ECN) */
	      pkt->lvl3.sflags[2] = isicmp & 0x04 ? 'R' : '*';	/* delay */
	      pkt->lvl3.sflags[1] = isicmp & 0x08 ? 'T' : '*';	/* throughput */
	      pkt->lvl3.sflags[0] = isicmp & 0x10 ? 'D' : '*';	/* reliability */
	      pkt->lvl3.pflags[3] = 0;	/* Null terminate precedence flags */
	      pkt->lvl3.pflags[2] = isicmp & 0x20 ? '3' : '*';	/* precedence */
	      pkt->lvl3.pflags[1] = isicmp & 0x40 ? '2' : '*';	/* bits */
	      pkt->lvl3.pflags[0] = isicmp & 0x80 ? '1' : '*';
	      break;
	    case 14:
	      if (strlen (out) >= sizeof pkt->lvl3.id)
		return (0);
	      strcpy (pkt->lvl3.id, out);
	      break;
	    case 15:
	      if (strlen (out) >= sizeof pkt->lvl3.frag)
		return (0);
	      strcpy (pkt->lvl3.frag, out);
	      sscanf (out, "0x%x", &isicmp);	/* isicmp is just a handy int */
	      pkt->lvl3.fflags[3] = 0;	/* Null terminate */
	      pkt->lvl3.fflags[2] = (isicmp >> 13) & 0x1 ? 'M' : '*';
	      pkt->lvl3.fflags[1] = (isicmp >> 13) & 0x2 ? 'D' : '*';
	      pkt->lvl3.fflags[0] = (isicmp >> 13) & 0x4 ? 'C' : '*';
	      break;
	    case 16:
	      if (strlen (out) >= sizeof pkt->lvl3.ttl)
		return (0);
	      strcpy (pkt->lvl3.ttl, out);
	      break;
	    case 17:
	      if (strlen (out) >= sizeof pkt->lvl4.pkt_TCP.flags)
		return (0);
	      strcpy (pkt->lvl4.pkt_TCP.flags, out);
	      if (strcmp (out, "SYN") == 0)
		strcpy (pkt->lvl4.pkt_TCP.tflags, "****S*");
	      break;
	    default:
	      continue;
	    }
	}
    }
  else
    return (0);
  free (matches);
  if (resolve_pkt (pkt))
    {
      return (1);
    }
  else
    return (0);
}
